package gov.va.cpss.dao.impl;

import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.springframework.jdbc.core.BatchPreparedStatementSetter;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.PreparedStatementCreator;
import org.springframework.jdbc.core.PreparedStatementSetter;
import org.springframework.jdbc.core.ResultSetExtractor;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;

import gov.va.cpss.cobol.Money;
import gov.va.cpss.dao.CBSSAuditDatesDAO;
import gov.va.cpss.dao.CBSStmtDAO;
import gov.va.cpss.model.BooleanChar;
import gov.va.cpss.model.cbs.CBSStmt;
import gov.va.cpss.model.printack.ADFileRecord;
import gov.va.cpss.model.printack.PrintAckRec;

/**
 * An implementation of the CBSStmtDAO interface.
 */
public class CBSStmtDAOImpl implements CBSStmtDAO {

	private static final String CBSSTMT_SELECT_FIELDS = "a.id AS id, a.replacedStmtId AS replacedStmtId, a.messageId AS messageId, "
			+ "a.accountId AS accountId, a.acntNumDisp AS acntNumDisp, a.amntDue AS amntDue, a.preBalance AS preBalance, "
			+ "a.totalCharges AS totalCharges, a.totalCredits AS totalCredits, a.newBalance AS newBalance, a.lateStmtMsg AS lateStmtMsg, "
			+ "a.statementDate AS statementDate, a.processDate AS processDate, a.batchRunId AS batchRunId, a.statusId AS statusId, "
			+ "a.printAckId AS printAckId, a.printDate AS printDate, "
			+ "a.createdBy AS createdBy, a.createdDate AS createdDate, a.modifiedBy AS modifiedBy, a.modifiedDate AS modifiedDate ";

	private final Logger daoLogger;

	private JdbcTemplate jdbcTemplate;

	private CBSSAuditDatesDAO cbssAuditDatesDAO;

	public CBSStmtDAOImpl() {
		daoLogger = Logger.getLogger(this.getClass().getCanonicalName());
	}

	public void setDataSource(DataSource dataSource) {
		this.jdbcTemplate = new JdbcTemplate(dataSource);
	}

	public CBSSAuditDatesDAO getCbssAuditDatesDAO() {
		return cbssAuditDatesDAO;
	}

	public void setCbssAuditDatesDAO(CBSSAuditDatesDAO cbssAuditDatesDAO) {
		this.cbssAuditDatesDAO = cbssAuditDatesDAO;
	}

	@Override
	public long save(CBSStmt statement) {
		// insert
		final String sql = "INSERT INTO CBSStmt (accountId, acntNumDisp, amntDue, preBalance, totalCharges, totalCredits, newBalance, lateStmtMsg, statementDate, processDate, batchRunId, statusId, replacedStmtId) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";
		KeyHolder keyHolder = new GeneratedKeyHolder();

		jdbcTemplate.update(new PreparedStatementCreator() {
			@Override
			public PreparedStatement createPreparedStatement(Connection connection) throws SQLException {
				PreparedStatement ps = connection.prepareStatement(sql, new String[] { "id" });
				ps.setLong(1, statement.getAccountId());
				ps.setString(2, statement.getAcntNumDisp());
				ps.setDouble(3, statement.getAmountDue().getDouble());
				ps.setDouble(4, statement.getPrevBalance().getDouble());
				ps.setDouble(5, statement.getTotalCharges().getDouble());
				ps.setDouble(6, statement.getTotalCredits().getDouble());
				ps.setDouble(7, statement.getNewBalance().getDouble());
				ps.setString(8, Character.toString(BooleanChar.from(statement.getLateStmtMsg()).getCharValue()));
				ps.setDate(9, new Date(statement.getStatementDate().getTime()));
				ps.setDate(10, new Date(statement.getProcessDate().getTime()));
				ps.setLong(11, statement.getBatchRunId());
				ps.setInt(12, statement.getStatusId());
				ps.setLong(13, statement.getReplacedStmtId());
				return ps;
			}
		}, keyHolder);
		statement.setId(keyHolder.getKey().longValue());

		// Load the new audit fields and set on statement
		cbssAuditDatesDAO.getAuditDates("CBSStmt", statement.getId(), statement);

		return statement.getId();
	}

	@Override
	public CBSStmt getExistingStatementByAccountAndMonth(final long accountNumber, final java.util.Date statementDate) {

		daoLogger.info("GETTING STATEMENT FOR DATE: |" + new SimpleDateFormat("MM-yyyy").format(statementDate) + "|");

		// query
		final String sql = "SELECT * FROM (SELECT * FROM CBSStmt where accountId=? and to_char(statementDate,'mm-yyyy')=? ORDER BY id DESC) WHERE rownum <= 1";

		CBSStmt results = jdbcTemplate.query(sql, new PreparedStatementSetter() {
			public void setValues(PreparedStatement preparedStatement) throws SQLException {
				preparedStatement.setLong(1, accountNumber);
				preparedStatement.setString(2, new SimpleDateFormat("MM-yyyy").format(statementDate));
			}
		}, new ResultSetExtractor<CBSStmt>() {
			@Override
			public CBSStmt extractData(ResultSet rs) throws SQLException {
				if (rs.next()) {
					CBSStmt statement = new CBSStmt(rs.getLong("batchRunId"), rs.getLong("accountId"),
							rs.getInt("statusId"));

					statement.setId(rs.getLong("id"));
					statement.setMessageId(rs.getLong("messageId"));
					statement.setAcntNumDisp(rs.getString("acntNumDisp"));
					statement.setAmountDue(new Money(rs.getDouble("amntDue"), 9));
					statement.setPrevBalance(new Money(rs.getDouble("preBalance"), 9));
					statement.setTotalCharges(new Money(rs.getDouble("totalCharges"), 9));
					statement.setTotalCredits(new Money(rs.getDouble("totalCredits"), 9));
					statement.setNewBalance(new Money(rs.getDouble("newBalance"), 9));
					statement.setLateStmtMsg(BooleanChar.from(rs.getString("lateStmtMsg")).isTrue());
					statement.setStatementDate(rs.getDate("statementDate"));
					statement.setProcessDate(rs.getDate("processDate"));
					statement.setReplacedStmtId(rs.getLong("replacedStmtId"));
					statement.setCreatedBy(rs.getString("createdBy"));
					statement.setCreatedDate(rs.getDate("createdDate"));
					statement.setModifiedBy(rs.getString("modifiedBy"));
					statement.setModifiedDate(rs.getDate("modifiedDate"));

					return statement;
				}
				return null;
			}
		});

		return results;
	}

	@Override
	public List<CBSStmt> getMessageStatementsForSite(final long cbsMessageId, final String stationNum) {
		final String sql = "SELECT " + CBSSTMT_SELECT_FIELDS
				+ "FROM CBSStmt a INNER JOIN CBSSiteStmt primarySite ON a.id = primarySite.stmtId "
				+ "WHERE a.messageId=" + cbsMessageId + " AND primarySite.isPrimary='Y' AND primarySite.stationNum='"
				+ stationNum + "' ";

		final List<CBSStmt> results = jdbcTemplate.query(sql, new CBSStmtRowMapper());

		return results;
	}

	@Override
	public List<CBSStmt> getStatementsForPrintAck(final PrintAckRec printAckRec, final java.util.Date statementDate,
			final int sentStatusId, final int ackStatusId) {

		final String formattedStatementDate = new SimpleDateFormat("MM-dd-yyyy").format(statementDate);

		final String sql = "SELECT " + CBSSTMT_SELECT_FIELDS
				+ "FROM CBSStmt a INNER JOIN CBSSiteStmt siteStmt ON a.id = siteStmt.stmtId "
				+ "  INNER JOIN CBSSitePatient patient ON siteStmt.id = patient.siteStmtId "
				+ "WHERE to_char(a.statementDate,'mm-dd-yyyy')='" + formattedStatementDate + "' "
				+ " AND a.statusId IN (" + sentStatusId + ", " + ackStatusId + ") AND siteStmt.stationNum='"
				+ printAckRec.getStationNum() + "' " + " AND patient.oldAcntNum=? ";
		daoLogger.debug(sql);

		final List<ADFileRecord> printAckDetailL = printAckRec.getPrintAckDetailL();

		final List<CBSStmt> cbsStmtL = new ArrayList<CBSStmt>((printAckDetailL == null) ? 0 : printAckDetailL.size());

		final CBSStmtRowMapper cbsStmtRowMapper = new CBSStmtRowMapper();

		if ((printAckDetailL != null) && (printAckDetailL.size() > 0)) {
			for (ADFileRecord adFileRecord : printAckDetailL) {
				CBSStmt patientStmt = null;

				if (adFileRecord != null) {
					final String patientAccount = adFileRecord.getPatientAccount();

					final List<CBSStmt> patientCbsStmtL = jdbcTemplate.query(sql, new PreparedStatementSetter() {
						public void setValues(PreparedStatement preparedStatement) throws SQLException {
							daoLogger.debug(patientAccount);
							preparedStatement.setString(1, patientAccount);
						}
					}, cbsStmtRowMapper);

					if ((patientCbsStmtL != null) && (patientCbsStmtL.size() > 0)) {
						patientStmt = patientCbsStmtL.get(0);
					}
				}
				cbsStmtL.add(patientStmt);
			}
		}

		return cbsStmtL;
	}

	@Override
	public CBSStmt get(long id) {
		final String sql = "SELECT * FROM CBSStmt WHERE id = " + id;

		final List<CBSStmt> cbsStmtL = jdbcTemplate.query(sql, new CBSStmtRowMapper());

		return (cbsStmtL.size() > 0) ? cbsStmtL.get(0) : null;
	}

	/**
	 * Maps a ResultSet to a new CBSStmt object.
	 * 
	 * @author Brad Pickle
	 *
	 */
	private static class CBSStmtRowMapper implements RowMapper<CBSStmt> {

		@Override
		public CBSStmt mapRow(ResultSet rs, int row) throws SQLException {
			final CBSStmt cbsStmt = new CBSStmt(rs.getLong("accountId"));

			cbsStmt.setId(rs.getLong("id"));
			cbsStmt.setReplacedStmtId(rs.getLong("replacedStmtId"));
			cbsStmt.setMessageId(rs.getLong("messageId"));
			cbsStmt.setAcntNumDisp(rs.getString("acntNumDisp"));
			cbsStmt.setAmountDue(new Money(rs.getDouble("amntDue"), 9));
			cbsStmt.setPrevBalance(new Money(rs.getDouble("preBalance"), 9));
			cbsStmt.setTotalCharges(new Money(rs.getDouble("totalCharges"), 9));
			cbsStmt.setTotalCredits(new Money(rs.getDouble("totalCredits"), 9));
			cbsStmt.setNewBalance(new Money(rs.getDouble("newBalance"), 9));
			cbsStmt.setLateStmtMsg(BooleanChar.from(rs.getString("lateStmtMsg")).isTrue());
			cbsStmt.setStatementDate(rs.getDate("statementDate"));
			cbsStmt.setProcessDate(rs.getDate("processDate"));
			cbsStmt.setBatchRunId(rs.getLong("batchRunId"));
			cbsStmt.setStatusId(rs.getInt("statusId"));
			cbsStmt.setPrintAckId(rs.getLong("printAckId"));
			cbsStmt.setPrintDate(rs.getDate("printDate"));
			cbsStmt.setCreatedBy(rs.getString("createdBy"));
			cbsStmt.setCreatedDate(rs.getDate("createdDate"));
			cbsStmt.setModifiedBy(rs.getString("modifiedBy"));
			cbsStmt.setModifiedDate(rs.getDate("modifiedDate"));

			return cbsStmt;
		}

	}

	@Override
	public Long getStatementCountWithStatus(int statusId) {
		final String sql = "SELECT COUNT(*) FROM CBSStmt WHERE statusId=" + statusId;

		final List<Long> results = jdbcTemplate.query(sql, new RowMapper<Long>() {
			@Override
			public Long mapRow(ResultSet rs, int row) throws SQLException {
				return new Long(rs.getLong(1));
			}
		});

		if ((results != null) && !results.isEmpty()) {
			return results.get(0);
		}

		return null;
	}

	@Override
	public int updateMessageIdForStatus(int statusId, long newMessageId) {
		final String sql = "UPDATE CBSStmt SET messageId=" + newMessageId + " WHERE statusId=" + statusId;
		return jdbcTemplate.update(sql);
	}

	@Override
	public int updateStatusForMessageId(int newStatusId, long messageId) {
		final String sql = "UPDATE CBSStmt SET statusId=" + newStatusId + " WHERE messageId=" + messageId;
		return jdbcTemplate.update(sql);
	}

	@Override
	public int updateAllStatusFromTo(final int fromStatus, final int toStatus) {
		final String sql = "UPDATE CBSStmt SET statusId=? WHERE statusId=?";
		return jdbcTemplate.update(sql, toStatus, fromStatus);
	}

	@Override
	public int updateStatusFromToForInitial(final int fromStatus, final int toStatus, final int initialStatus) {
		final String sql = "UPDATE CBSStmt SET statusId=? WHERE statusId=? and id in (select replacedstmtid from CBSStmt where statusid=? and replacedstmtid not in 0)";
		return jdbcTemplate.update(sql, toStatus, fromStatus, initialStatus);
	}

	@Override
	public void updatePrintAckFields(final List<CBSStmt> cbsStmtL, final long printAckId,
			final java.util.Date printDate, final int statusId, final String stationNum) {
		final String formattedPrintDate = new SimpleDateFormat("MM-dd-yyyy").format(printDate);

		final String sql = "UPDATE CBSStmt SET printAckId=" + printAckId + ", printDate=TO_DATE('" + formattedPrintDate
				+ "', 'MM-DD-YYYY'), statusId=" + statusId
				+ " WHERE id=? AND ((printAckId=null OR printDate=null OR statusId!=" + statusId
				+ ") OR id=(SELECT stmtId FROM CBSSiteStmt WHERE stmtId=CBSStmt.id AND isPrimary='Y' AND stationNum='"
				+ stationNum + "'))";

		jdbcTemplate.batchUpdate(sql, new BatchPreparedStatementSetter() {

			@Override
			public void setValues(PreparedStatement ps, int i) throws SQLException {
				ps.setLong(1, cbsStmtL.get(i).getId());
			}

			@Override
			public int getBatchSize() {
				return cbsStmtL.size();
			}
		});
	}

	@Override
	public int[] undoPrintAcknowledgements(final List<Long> printAckRecIdList, final int sentStatusId) {
		if ((printAckRecIdList == null) || (printAckRecIdList.size() == 0))
			return new int[0];

		final String sql = "UPDATE CBSStmt SET statusId=" + sentStatusId
				+ ", printDate=NULL, printAckId=NULL WHERE printAckId=? ";

		final int[] rowCounts = new int[printAckRecIdList.size()];
		for (int i = 0; i < printAckRecIdList.size(); i++) {
			rowCounts[i] = jdbcTemplate.update(sql, printAckRecIdList.get(i));
		}

		return rowCounts;
	}

	@Override
	public int updateStatusFromToForPrintAckId(final int fromStatus, final int toStatus, final long printAckId) {
		final String sql = "UPDATE CBSStmt SET statusId=? WHERE statusId=? and printAckId=? ";
		return jdbcTemplate.update(sql, toStatus, fromStatus, printAckId);
	}

	@Override
	public int deleteByStatusId(final int statusId) {

		final String sql = "DELETE FROM CBSStmt where statusId = ?";
		return jdbcTemplate.update(sql, statusId);
	}

	public static String getUpdateCBSStmtSql() {
		return "update CBSStmt set statusId=? where id=?";
	}

	public static String getCBSSStmtJoinSql() {
		return "SELECT DISTINCT CBSSiteStmt.stationNum AS stationNum "
				+ "FROM CBSStmt INNER JOIN CBSSiteStmt ON CBSStmt.id = CBSSiteStmt.stmtId "
				+ "WHERE CBSSiteStmt.isPrimary = 'Y' and CBSStmt.messageId = ? and CBSStmt.statusId = ?";
	}

}
